//
//
//	3d2 files parser class
//
//											(C) 1997 JoOl

#ifndef _PARSER_IOB_H_
#define _PARSER_IOB_H_

#include "ParserBin.h"


class						ParserIob : public ParserBin
{
	const char*				QuickDraw[3];
	const char*				PartShape[10];
	const char*				PartCenter[4];
	const char*				PartSize[4];
	const char*				PartAlign[3];

	bool					SkipChunk(uint32);
	bool					ParseColor(uchar& , uchar& , uchar& );
	bool					ParseColorPad(uchar& , uchar& , uchar& );
	bool					ParseFract(float& );
	bool					ParseVector(float& , float& , float& );
	bool					ParseMatrix();
	bool					ParseTform(uint32);

#define SKIP_CHUNK			PRINT6("skip chunk '%c%c%c%c' %d, cumul=%d\n",		\
								((ChunkId >> 24) & 0x00ff),						\
								((ChunkId >> 16) & 0x00ff),						\
								((ChunkId >> 8) & 0x00ff),						\
								(ChunkId & 0x00ff),								\
								ChunkLen, CurrLen);								\
							ASSERT(SkipChunk(ChunkLen))

#define PAD_CHUNK			if (Len & 0x01)							\
							{										\
								ASSERT(SkipBytes(1));				\
							}

#define PRINT_CHUNK			PRINT5("chunk '%c%c%c%c' %d\n",		\
								((ChunkId >> 24) & 0x00ff),			\
								((ChunkId >> 16) & 0x00ff),			\
								((ChunkId >> 8) & 0x00ff),			\
								(ChunkId & 0x00ff),					\
								ChunkLen);

#define CHECK(a)			if (CurrLen != Len)									\
							{													\
								PRINT2("@@@@@ " a ": %d/%d\n", CurrLen, Len);	\
								return false;									\
							}													\
							PRINT2(a ": %d/%d\n", CurrLen, Len);				\
							return true

#define MAX_NAMELEN			18


	bool					ParseHeader(uint32& );
	bool					ParseObject(uint32 );
	bool					ParseObjectDesc(uint32 );

	bool					ParseName(uint32, char* );

	bool					ParseShp2(uint32);
		void					PrintShp2Shape(uint16);
		void					PrintShp2Lamp(uint16);
	bool					ParsePosi(uint32);
	bool					ParseAxis(uint32);
	bool					ParseSize(uint32);
	bool					ParseBbox(uint32);
	bool					ParseStnd(uint32);
		bool					ParseStid(uint32);
			void					PrintStidFlags(uint16);
		bool					ParseStdt(uint32);
			bool					ParseStdtAxis(uint32);
			bool					ParseStdtVertex(uint32);
			bool					ParseStdtPath(uint32);
			bool					ParseStdtAxisSize(uint32);
			bool					ParseStdtFaceColor(uint32);
			bool					ParseStdtFaceFilter(uint32);
			bool					ParseStdtReflect(uint32);
			bool					ParseStdtObjAttrib(uint32);
			void					PrintPrp1(uchar* );
	bool					ParsePnts(uint32);
	bool					ParseEdge(uint32);
	bool					ParseFace(uint32);
	bool					ParsePth2(uint32);
		bool					ParsePthd(uint32);
	bool					ParseColr(uint32);
	bool					ParseRefl(uint32);
	bool					ParseTran(uint32);
	bool					ParseSpc1(uint32);
	bool					ParseClst(uint32);
	bool					ParseRlst(uint32);
	bool					ParseTlst(uint32);
	bool					ParseTxt3(uint32);
	bool					ParseBrs4(uint32);
		void					PrintBrs4Flags(uint16);
		void					PrintBrs4Wflags(uint16);
	bool					ParseFogl(uint32);
	bool					ParsePrp1(uint32);
	bool					ParseInt1(uint32);
	bool					ParseAnid(uint32);
	bool					ParseFor2(uint32);
		void					PrintFor2Flags(uint16, uint16& );
	bool					ParsePart(uint32);
		void					PrintPartType(uint16);
	bool					ParsePtfn(uint32);
	bool					ParseFgr2(uint32);
	bool					ParseBbsg(uint32);
	bool					ParseSbsg(uint32);
	bool					ParseEflg(uint32);

	enum
	{
		FORM_ID						= 'FORM',
		FORM_TYPE					= 'TDDD',

		OBJECT_ID					= 'OBJ ',

			// Sub chunks of OBJ
			OBJ_DESC				= 'DESC',
				OBJ_NAME			= 'NAME',
				OBJ_SHP2			= 'SHP2',
				OBJ_POSI			= 'POSI',
				OBJ_AXIS			= 'AXIS',
				OBJ_SIZE			= 'SIZE',
				OBJ_BBOX			= 'BBOX',
				OBJ_STND			= 'STND',
					OBJ_STID		= 'STID',
					OBJ_STDT		= 'STDT',
				OBJ_PNTS			= 'PNTS',
				OBJ_EDGE			= 'EDGE',
				OBJ_FACE			= 'FACE',
				OBJ_PTH2			= 'PTH2',

				OBJ_COLR			= 'COLR',
				OBJ_REFL			= 'REFL',
				OBJ_TRAN			= 'TRAN',
				OBJ_SPC1			= 'SPC1',

				OBJ_CLST			= 'CLST',
				OBJ_RLST			= 'RLST',
				OBJ_TLST			= 'RLST',

				OBJ_TXT3			= 'TXT3',
				OBJ_BRS4			= 'BRS4',
				OBJ_FOGL			= 'FOGL',
				OBJ_PRP1			= 'PRP1',
				OBJ_INT1			= 'INT1',
				OBJ_ANID			= 'ANID',
				OBJ_FOR2			= 'FOR2',
				OBJ_PART			= 'PART',
				OBJ_PTFN			= 'PTFN',
				OBJ_FGR2			= 'FGR2',
				OBJ_BBSG			= 'BBSG',
				OBJ_SBSG			= 'SBSG',
				OBJ_EFLG			= 'EFLG',
			OBJ_END					= 'TOBJ',
	};


	// OBJ_SHP2-related values
	enum
	{
		SHP2_TYPE_SPHERE			= 0,
		SHP2_TYPE_AXIS				= 2,
		SHP2_TYPE_GROUND			= 5,

		SHP2_LAMP_POINT_SOURCE		= BIT00,
		SHP2_LAMP_PARALLEL_SOURCE	= BIT01,
		SHP2_LAMP_ROUND_SHAPE		= BIT02,
		SHP2_LAMP_RECT_SHAPE		= BIT03,
		SHP2_LAMP_LENS_FLARE		= BIT04,
		SHP2_LAMP_DIMINISH_I		= BIT05,
		SHP2_LAMP_FALLOFF			= BIT06,
		SHP2_LAMP_CAST_SHADOWS		= BIT07,
		SHP2_LAMP_BRIGHT_OBJ		= BIT15
	};

	enum
	{
		STID_AXIS_DATA				= BIT00,
		STID_SHAPE_DATA				= BIT01,
		STID_COLOR_DATA				= BIT02,

		STDT_AXIS_DATA				= 101,
		STDT_VERTEX_DATA			= 102,
		STDT_PATH_DATA				= 103,
		STDT_AXIS_SIZE_DATA			= 104,
		STDT_FACE_COLOR_DATA		= 105,
		STDT_FACE_FILTER_DATA		= 106,
		STDT_FACE_REFLECT_DATA		= 107,
		STDT_OBJ_ATTRIB_DATA		= 108,
	};

	enum
	{
		TXT3_APPLY_TO_CHILDREN		= BIT00,

		TXT3_FLAG_ALTER_RED			= BIT00,
		TXT3_FLAG_ALTER_GREEN		= BIT01,
		TXT3_FLAG_ALTER_BLUE		= BIT02,
		TXT3_FLAG_ALTER_PARAM		= BIT03,
	};

	enum
	{
		BRS4_FLAG_COLOR				= BIT00,
		BRS4_FLAG_REFLECTIVITY		= BIT01,
		BRS4_FLAG_FILTER			= BIT02,
		BRS4_FLAG_ALTITUDE			= BIT03,
		BRS4_FLAG_REFLECTION		= BIT04,

		BRS4_WFLAG_WRAP_X					= BIT00,
		BRS4_WFLAG_WRAP_Z					= BIT01,
		BRS4_WFLAG_WRAP_APPLY_TO_CHILDREN	= BIT02,
		BRS4_WFLAG_WRAP_REPEAT				= BIT03,
		BRS4_WFLAG_WRAP_MIRROR				= BIT04,
		BRS4_WFLAG_WRAP_INVERSE_VIDEO		= BIT05,
		BRS4_WFLAG_WRAP_USE_GENLOCK			= BIT06,
	};


	enum
	{
		FOR2_XY_CROSS_SECTION		= BIT00,
		FOR2_ONE_FORMER_VIEW		= BIT01,
		FOR2_ONE_SPACER_VIEW		= BIT02,
	};

	enum
	{
		PART_SHAPE_FACE				= 0,
		PART_SHAPE_TETRAHEDRON		= 1,
		PART_SHAPE_PYRAMID			= 2,
		PART_SHAPE_OCTAHEDRON		= 3,
		PART_SHAPE_CUBE				= 4,
		PART_SHAPE_BLOCK			= 5,
		PART_SHAPE_DODECAHEDRON		= 6,
		PART_SHAPE_SPHERE			= 7,
		PART_SHAPE_RANDOM			= 8,
		PART_SHAPE_USE_OBJECT_FILE	= 9,

		PART_CENTER_INSCRIBED		= 0x0000,
		PART_CENTER_CIRCUMSCRIBED	= 0x0010,
		PART_CENTER_INTERPOLATED	= 0x0020,
		PART_CENTER_BARYCENTRIC		= 0x0030,

		PART_SIZE_SMALL				= 0x0000,
		PART_SIZE_LARGE				= 0x0100,
		PART_SIZE_RANDOM			= 0x0200,
		PART_SIZE_SPECIFY			= 0x0300,

		PART_ALIGNMENT_OBJECT		= 0x1000,
		PART_ALIGNMENT_FACE			= 0x2000,
		PART_ALIGNMENT_RANDOM		= 0x3000,
	};

	enum
	{
		EFLG_QUICK_EDGE				= 0x0040,
		EFLG_SHARP_EDGE				= 0x0080,
	};


#define READ_CHUNKID		uint32			ChunkId;	\
							READ_INT32(ChunkId)

#define READ_CHUNKLEN		uint32			ChunkLen;	\
							READ_INT32(ChunkLen)

#define READ_CHUNK			READ_CHUNKID;				\
							READ_CHUNKLEN



public:

	ParserIob(bool = false, int32 = LEXER_BUFLEN, int32 = LEXER_LEXMLEN);
	~ParserIob();

	virtual bool			SetInput(entry_ref& );

	virtual bool			CheckType();
	virtual bool			Parse();
};


#endif	/* _PARSER_IOB_H_ */



