//
//
//	OBJ file parser class
//
//											(C) 1997 JoOl


#include "ParserObj.h"



ParserObj::ParserObj(bool V, int32 BufLen, int32 LexmLen): ParserText(V)
{
	Lexr = new LexerObj(BufLen, LexmLen);
}


ParserObj::~ParserObj()
{
	delete Lexr;
}


bool
ParserObj::CheckType()
{
	int32		Cpt = 10;
	while ((!Peek(LX_CHAR_EOF)) && (Cpt--))
	{
		ASSERT(ParseSwitch());
	}

	return true;
}


bool
ParserObj::ParseVector1or2or3(float& x, float& y, float& z)
{
	ASSERT(ParseFloat(x));
	if (Peek(LX_TYPE_FLOAT) || Peek(LX_TYPE_INTEGER))
	{
		ASSERT(ParseFloat(y));
	}
	else
		y = 0.0;
	if (Peek(LX_TYPE_FLOAT) || Peek(LX_TYPE_INTEGER))
	{
		ASSERT(ParseFloat(z));
	}
	else
		z = 0.0;
	return true;
}


bool
ParserObj::ParseVector3(float& x, float& y, float& z)
{
	ASSERT(ParseFloat(x));
	ASSERT(ParseFloat(y));
	ASSERT(ParseFloat(z));
	return true;
}


bool
ParserObj::ParseVector3or4(float& x, float& y, float& z, float& w)
{
	ASSERT(ParseFloat(x));
	ASSERT(ParseFloat(y));
	ASSERT(ParseFloat(z));
	if (Peek(LX_TYPE_FLOAT) || Peek(LX_TYPE_INTEGER))
	{
		ASSERT(ParseFloat(w));
	}
	else
		w = 1.0;

	return true;
}


bool
ParserObj::ParseTriplet(int32& v, int32& vt, int32& vn)
{
	ASSERT(LexInt(v));
	if (Peek(LX_CHAR_SLASH))
	{
		MATCH(LX_CHAR_SLASH);
		if (Peek(LX_TYPE_INTEGER))
		{
			ASSERT(LexInt(vt));
		}
		else
			vt = INT32_NO_VALUE;
		if (Peek(LX_CHAR_SLASH))
		{
			MATCH(LX_CHAR_SLASH);
			ASSERT(LexInt(vn));
		}
		else
			vn = INT32_NO_VALUE;
	}
	else
	{
		vt = vn = INT32_NO_VALUE;
	}

	return true;
}


void
ParserObj::PrintTriplet(int32 IndexVert, int32 IndexTexture,
						int32 IndexNormal)
{
	_PRINT1("[%d, ", IndexVert);
	if (uint32(IndexTexture) != INT32_NO_VALUE)
		_PRINT1("%d", IndexTexture);
	_PRINT0(", ");
	if (uint32(IndexNormal) != INT32_NO_VALUE)
		_PRINT1("%d", IndexNormal);
	_PRINT0("] ");

}


bool
ParserObj::ParseVertex()
{
	float		x, y, z, w;

	MATCH(LX_TOKEN_V);
	ASSERT(ParseVector3or4(x, y, z, w));
	PRINT4("v %f %f %f %f\n", x, y, z, w);

	return true;
}


bool
ParserObj::ParseVtexture()
{
	float		x, y, z;

	MATCH(LX_TOKEN_VTEXTURE);
	ASSERT(ParseVector1or2or3(x, y, z));
	PRINT3("vt %f %f %f\n", x, y, z);

	return true;
}


bool
ParserObj::ParseVnormal()
{
	float		x, y, z;

	MATCH(LX_TOKEN_VNORMAL);
	ASSERT(ParseVector3(x, y, z));
	PRINT3("vn %f %f %f\n", x, y, z);

	return true;
}


bool
ParserObj::ParseVparam()
{
	float		x, y, z;

	MATCH(LX_TOKEN_VPARAM);
	ASSERT(ParseVector1or2or3(x, y, z));
	PRINT3("vp %f %f %f\n", x, y, z);

	return true;
}


bool
ParserObj::ParseSwitch()
{
	switch (ParserToken)
	{
		case LX_TOKEN_V:				ASSERT(ParseVertex());					break;
		case LX_TOKEN_VTEXTURE:			ASSERT(ParseVtexture());				break;
		case LX_TOKEN_VNORMAL:			ASSERT(ParseVnormal());					break;
		case LX_TOKEN_VPARAM:			ASSERT(ParseVparam());					break;
		case LX_TOKEN_CSTYPE:			ASSERT(ParseCurveOrSurface());			break;
		case LX_TOKEN_CTECH:			ASSERT(ParseCtech());					break;
		case LX_TOKEN_STECH:			ASSERT(ParseStech());					break;
		case LX_TOKEN_DEGREE:			ASSERT(ParseCsDegree(DegU, DegV));		break;
		case LX_TOKEN_BMATU:			PRINT0("bmat u:\n");
										MATCH(LX_TOKEN_BMATU);
										ASSERT(ParseMatrix(DegU+1));			break;
		case LX_TOKEN_BMATV:			PRINT0("bmat v:\n");
										MATCH(LX_TOKEN_BMATV);
										ASSERT(ParseMatrix(DegV+1));			break;
		case LX_TOKEN_STEP:				ASSERT(ParseSteps(StepU, StepV));		break;
		case LX_TOKEN_POINT:			ASSERT(ParsePoint());					break;
		case LX_TOKEN_LINE:				ASSERT(ParseLine());					break;
		case LX_TOKEN_FACE:				ASSERT(ParseFace());					break;
		case LX_TOKEN_CURVE:			ASSERT(ParseCurve());					break;
		case LX_TOKEN_CURVE2D:			ASSERT(ParseCurve2d());					break;
		case LX_TOKEN_SURFACE:			ASSERT(ParseSurface());					break;
		case LX_TOKEN_PARAMU:			ASSERT(ParseParam());					break;
		case LX_TOKEN_PARAMV:			ASSERT(ParseParam());					break;
		case LX_TOKEN_TRIM:				ASSERT(ParseTrim());					break;
		case LX_TOKEN_HOLE:				ASSERT(ParseHole());					break;
		case LX_TOKEN_SCURVE:			ASSERT(ParseScurve());					break;
		case LX_TOKEN_SPOINT:			ASSERT(ParseSpoints());					break;
		case LX_TOKEN_CONNECT:			ASSERT(ParseConnect());					break;
		case LX_TOKEN_ENDSTMT:			MATCH(LX_TOKEN_ENDSTMT);				break;
		case LX_TOKEN_GROUPNAME:		ASSERT(ParseGroup());					break;
		case LX_TOKEN_SMOOTHGROUP:		ASSERT(ParseSmoothGroup());				break;
		case LX_TOKEN_MERGEGROUP:		ASSERT(ParseMergeGroup());				break;
		case LX_TOKEN_OBJECTNAME:		ASSERT(ParseObjectName());				break;
		case LX_TOKEN_BEVELINTERP:		ASSERT(ParseBevelInterpol());			break;
		case LX_TOKEN_CINTERP:			ASSERT(ParseCinterp());					break;
		case LX_TOKEN_DINTERP:			ASSERT(ParseDinterp());					break;
		case LX_TOKEN_LOD:				ASSERT(ParseLod());						break;
		case LX_TOKEN_MAPLIB:			ASSERT(ParseMapLib());					break;
		case LX_TOKEN_USEMAP:			ASSERT(ParseUseMap());					break;
		case LX_TOKEN_USEMATERIAL:		ASSERT(ParseUseMaterial());				break;
		case LX_TOKEN_MATLIB:			ASSERT(ParseMaterialLib());				break;
		case LX_TOKEN_SHADOW:			ASSERT(ParseShadow());					break;
		case LX_TOKEN_TRACE:			ASSERT(ParseTrace());					break;
		case LX_TOKEN_BSP:				ASSERT(ParseBsp());						break;
		case LX_TOKEN_BZP:				ASSERT(ParseBzp());						break;
		case LX_TOKEN_CDC:				ASSERT(ParseCdc());						break;
		case LX_TOKEN_CDP:				ASSERT(ParseCdp());						break;
		case LX_TOKEN_RES:				ASSERT(ParseRes());						break;
		default:
			PRINT2("*** token %d = %s: skip line\n", ParserToken, Lexr->GetTokenLexeme(ParserToken));
			return false;
	}
	return true;
}



//----------------------------------------
// Main parsing routine
//----------------------------------------


bool
ParserObj::Parse()
{
	while (!Peek(LX_CHAR_EOF))
	{
		ASSERT(ParseSwitch());
	}

	Terminate();

	printf("Done...\n");
	return true;
}


//----------------------------------------
// Curve or Surface-related routines
//----------------------------------------

bool
ParserObj::ParseCurveOrSurface()
{
	bool			Rational = false;
	int32			CsType;

	MATCH(LX_TOKEN_CSTYPE);

	if (Peek(LX_TOKEN_RAT))
	{
		MATCH(LX_TOKEN_RAT);
		Rational = true;
	}
	ASSERT2(ParseCsType(CsType), "cstype");
	PRINT2("cstype %s rational, type=%s\n", Rational?"":"non", Lexr->GetTokenLexeme(CsType));

	return true;
}


bool
ParserObj::ParseCsType(int32& Out)
{
	if (Peek(LX_TOKEN_BMATRIX))
	{
		MATCH(LX_TOKEN_BMATRIX);
		Out = LX_TOKEN_BMATRIX;
	}
	else if (Peek(LX_TOKEN_BEZIER))
	{
		MATCH(LX_TOKEN_BEZIER);
		Out = LX_TOKEN_BEZIER;
	}
	else if (Peek(LX_TOKEN_BSPLINE))
	{
		MATCH(LX_TOKEN_BSPLINE);
		Out = LX_TOKEN_BSPLINE;
	}
	else if (Peek(LX_TOKEN_CARDINAL))
	{
		MATCH(LX_TOKEN_CARDINAL);
		Out = LX_TOKEN_CARDINAL;
	}
	else if (Peek(LX_TOKEN_TAYLOR))
	{
		MATCH(LX_TOKEN_TAYLOR);
		Out = LX_TOKEN_TAYLOR;
	}
	else
		return false;

	return true;
}


bool
ParserObj::ParseCsDegree(int32& DegU, int32& DegV)
{
	MATCH(LX_TOKEN_DEGREE);

	ASSERT(LexInt(DegU));
	if (Peek(LX_TYPE_INTEGER))
	{
		ASSERT(LexInt(DegV));
	}
	else
		DegV = 0;
	PRINT2("degree u=%d v=%d\n", DegU, DegV);

	return true;
}


bool
ParserObj::ParseMatrix(int32 Deg)
{
	for (int32 i = 0; i < Deg; i++)
	{
		float		Mij;
		PRINT0("mat ");
		for (int32 j = 0; j < Deg; j++)
		{
			ASSERT(ParseFloat(Mij));
			_PRINT1("%f ", Mij);
		}
		_PRINT0("\n");
	}

	return true;
}



bool
ParserObj::ParseSteps(int32& StepU, int32& StepV)
{
	MATCH(LX_TOKEN_STEP);

	ASSERT(LexInt(StepU));
	if (Peek(LX_TYPE_INTEGER))
	{
		ASSERT(LexInt(StepV));
	}
	else
		StepV = 0;
	PRINT2("step u=%d, v=%d\n", StepU, StepV);

	return true;
}


bool
ParserObj::ParseParam()
{
	float				p;

	if (Peek(LX_TOKEN_PARAMU))
	{
		MATCH(LX_TOKEN_PARAMU);
		PRINT0("param u: ");
	}
	else
	{
		MATCH(LX_TOKEN_PARAMV);
		PRINT0("param v: ");
	}

	while (Peek(LX_TYPE_FLOAT))
	{
		ASSERT(ParseFloat(p));
		_PRINT1("%f ", p);
	}
	_PRINT0("\n");


	return true;
}


bool
ParserObj::ParseTrim()
{
	MATCH(LX_TOKEN_TRIM);
	PRINT0("trim: ");

	while (Peek(LX_TYPE_FLOAT))
	{
		float			u0, u1;
		int32			CurvIndex;
		ASSERT(ParseFloat(u0));
		ASSERT(ParseFloat(u1));
		ASSERT(LexInt(CurvIndex));

		_PRINT3("[%f %f %d] ", u0, u1, CurvIndex);
	}
	_PRINT0("\n");

	return true;
}


bool
ParserObj::ParseHole()
{
	MATCH(LX_TOKEN_HOLE);
	PRINT0("hole: ");

	while (Peek(LX_TYPE_FLOAT))
	{
		int32			CurvIndex;
		float			u0, u1;
		ASSERT(ParseFloat(u0));
		ASSERT(ParseFloat(u1));
		ASSERT(LexInt(CurvIndex));

		_PRINT3("[%f %f %d] ", u0, u1, CurvIndex);
	}
	_PRINT0("\n");

	return true;
}


bool
ParserObj::ParseScurve()
{
	MATCH(LX_TOKEN_SCURVE);
	PRINT0("scurve: ");

	while (Peek(LX_TYPE_FLOAT))
	{
		int32			CurvIndex;
		float			u0, u1;
		ASSERT(ParseFloat(u0));
		ASSERT(ParseFloat(u1));
		ASSERT(LexInt(CurvIndex));

		_PRINT3("[%f %f %d] ", u0, u1, CurvIndex);
	}
	_PRINT0("\n");

	return true;
}


bool
ParserObj::ParseSpoints()
{
	MATCH(LX_TOKEN_SPOINT);
	PRINT0("special points: ");

	while (Peek(LX_TYPE_INTEGER))
	{
		int32			VertIndex;
		ASSERT(LexInt(VertIndex));
		_PRINT1("%d ", VertIndex);
	}
	_PRINT0("\n");

	return true;
}


bool
ParserObj::ParseConnect()
{
	int32			Surf0, Surf1;
	float			q00, q01;
	float			q10, q11;
	int32			Curv0, Curv1;
	MATCH(LX_TOKEN_CONNECT);

	ASSERT(LexInt(Surf0));
	ASSERT(ParseFloat(q00));
	ASSERT(ParseFloat(q01));
	ASSERT(LexInt(Curv0));

	ASSERT(LexInt(Surf1));
	ASSERT(ParseFloat(q10));
	ASSERT(ParseFloat(q11));
	ASSERT(LexInt(Curv1));

	PRINT8("connect from surf %d %f %f, curve %d to %d: %f %f, curve %d\n",
			Surf0, q00, q01, Curv0, Surf1, q10, q11, Curv1);
	return true;
}


//----------------------------------------
// Elements-related routines
//----------------------------------------

bool
ParserObj::ParsePoint()
{
	MATCH(LX_TOKEN_POINT);
	PRINT0("points: ");

	while (Peek(LX_TYPE_INTEGER))
	{
		int32			IndexVert;
		int32			IndexTexture;

		ASSERT(LexInt(IndexVert));
		if (Peek(LX_CHAR_SLASH))
		{
			MATCH(LX_CHAR_SLASH);
			ASSERT(LexInt(IndexTexture));
			_PRINT2("[%d, %d, ] ", IndexVert, IndexTexture);
		}
		else
		{
			_PRINT1("[%d, , ] ", IndexVert);
		}
	}
	_PRINT0("\n");

	return true;
}



bool
ParserObj::ParseLine()
{
	MATCH(LX_TOKEN_LINE);
	PRINT0("lines: ");

	while (Peek(LX_TYPE_INTEGER))
	{
		int32			IndexVert;
		int32			IndexTexture;

		ASSERT(LexInt(IndexVert));
		if (Peek(LX_CHAR_SLASH))
		{
			MATCH(LX_CHAR_SLASH);
			ASSERT(LexInt(IndexTexture));
			_PRINT2("[%d, %d, ] ", IndexVert, IndexTexture);
		}
		else
		{
			_PRINT1("[%d, , ] ", IndexVert);
		}
	}
	_PRINT0("\n");

	return true;
}



bool
ParserObj::ParseFace()
{
	int32				NbVert = 0;

	MATCH(LX_TOKEN_FACE);
	PRINT0("face: ");

	while (Peek(LX_TYPE_INTEGER))
	{
		int32			IndexVert;
		int32			IndexTexture;
		int32			IndexNormal;

		ASSERT(ParseTriplet(IndexVert, IndexTexture, IndexNormal));
		PrintTriplet(IndexVert, IndexTexture, IndexNormal);

		NbVert++;
	}
	_PRINT1(" => %d vertices\n", NbVert);

	return true;
}


bool
ParserObj::ParseCurve()
{
	float			u0, u1;

	MATCH(LX_TOKEN_CURVE);
	ASSERT(ParseFloat(u0));
	ASSERT(ParseFloat(u1));
	PRINT2("curve from %f to %f: ", u0, u1);

	while (Peek(LX_TYPE_INTEGER))
	{
		int32		Vert;
		ASSERT(LexInt(Vert));
		_PRINT1("%d ", Vert);
	}
	_PRINT0("\n");

	return true;
}


bool
ParserObj::ParseCurve2d()
{
	MATCH(LX_TOKEN_CURVE2D);
	PRINT0("curve 2d: vp= ");

	while (Peek(LX_TYPE_INTEGER))
	{
		int32		Vert;
		ASSERT(LexInt(Vert));
		_PRINT1("%d ", Vert);
	}
	_PRINT0("\n");

	return true;
}


bool
ParserObj::ParseSurface()
{
	float			s0, s1;
	float			t0, t1;

	MATCH(LX_TOKEN_SURFACE);
	ASSERT(ParseFloat(s0));
	ASSERT(ParseFloat(s1));
	ASSERT(ParseFloat(t0));
	ASSERT(ParseFloat(t1));
	PRINT4("surf from s:%f -> %f, t=%f -> %f: ", s0, s1, t0, t1);

	while (Peek(LX_TYPE_INTEGER))
	{
		int32			IndexVert;
		int32			IndexTexture;
		int32			IndexNormal;

		ASSERT(ParseTriplet(IndexVert, IndexTexture, IndexNormal));
		PrintTriplet(IndexVert, IndexTexture, IndexNormal);
	}
	_PRINT0("\n");

	return true;
}


//----------------------------------------
// Grouping-related routines
//----------------------------------------


bool
ParserObj::ParseGroup()
{
	MATCH(LX_TOKEN_GROUPNAME);
	PRINT0("group: ");

	int32		Ln = Lexr->GetLineNumber();
	ASSERT(Lexr->SkipSpaces());
	while (Lexr->GetLineNumber() == Ln)
	{
		char*		GroupName;
		ASSERT(ParseWord(GroupName));
		_PRINT1("'%s' ", GroupName);
		ASSERT(Lexr->SkipSpaces());
	}
	_PRINT0("\n");

	return true;
}


bool
ParserObj::ParseSmoothGroup()
{
	MATCH(LX_TOKEN_SMOOTHGROUP);

	if (Peek(LX_TYPE_STRING))
	{
		char*			Off;
		ASSERT(ParseString(Off));
		if (!strcmp(Off, "off"))
		{
			PRINT0("smooth group: off\n");
		}
		else
		{
			PRINT1("smooth group: '%s' error?\n", Off);
		}
		return true;
	}

	PRINT0("smooth group: ");
	while (Peek(LX_TYPE_INTEGER))
	{
		int32			GroupIndex;
		ASSERT(LexInt(GroupIndex));
		_PRINT1("%d ", GroupIndex);
	}

	_PRINT0("\n");

	return true;
}


bool
ParserObj::ParseMergeGroup()
{
	int32			GroupIndex;

	MATCH(LX_TOKEN_MERGEGROUP);
	ASSERT(LexInt(GroupIndex));
	PRINT1("merge group %d", GroupIndex);
	if (Peek(LX_TYPE_FLOAT))
	{
		float			Reso;
		ASSERT(ParseFloat(Reso));
		_PRINT1(" reso %f\n", Reso);
	}
	_PRINT0("\n");

	return true;
}


bool
ParserObj::ParseObjectName()
{
	char*			ObjName;
	MATCH(LX_TOKEN_OBJECTNAME);

	ASSERT(Lexr->SkipSpaces());
	ASSERT(ParseWord(ObjName));
	PRINT1("object '%s'\n", ObjName);

	return true;
}


//----------------------------------------
// Rendering attributes related routines
//----------------------------------------


bool
ParserObj::ParseBevelInterpol()
{
	MATCH(LX_TOKEN_BEVELINTERP);
	char*			OnOff;
	ASSERT(ParseString(OnOff));
	if (!strcmp(OnOff, "on"))
	{
		PRINT0("bevel on\n");
	}
	else
	{
		PRINT0("bevel off\n");
	}

	return true;
}


bool
ParserObj::ParseCinterp()
{
	MATCH(LX_TOKEN_CINTERP);
	char*			OnOff;
	ASSERT(ParseString(OnOff));
	if (!strcmp(OnOff, "on"))
	{
		PRINT0("cinterp on\n");
	}
	else
	{
		PRINT0("cinterp off\n");
	}

	return true;
}


bool
ParserObj::ParseDinterp()
{
	MATCH(LX_TOKEN_DINTERP);
	char*			OnOff;
	ASSERT(ParseString(OnOff));
	if (!strcmp(OnOff, "on"))
	{
		PRINT0("dinterp on\n");
	}
	else
	{
		PRINT0("dinterp off\n");
	}

	return true;
}


bool
ParserObj::ParseLod()
{
	int32			Lod;
	MATCH(LX_TOKEN_LOD);
	ASSERT(LexInt(Lod));
	PRINT1("lod %d\n", Lod);

	return true;
}


bool
ParserObj::ParseMapLib()
{
	MATCH(LX_TOKEN_MAPLIB);
	PRINT0("maplib: ");

	int32		Ln = Lexr->GetLineNumber();
	ASSERT(Lexr->SkipSpaces());
	while (Lexr->GetLineNumber() == Ln)
	{
		char*		MapFile;
		ASSERT(ParseWord(MapFile));
		_PRINT1("'%s' ", MapFile);
		ASSERT(Lexr->SkipSpaces());
	}
	_PRINT0("\n");


	return true;
}


bool
ParserObj::ParseUseMap()
{
	char*			MapName;
	MATCH(LX_TOKEN_USEMAP);
	ASSERT(Lexr->SkipSpaces());
	ASSERT(ParseWord(MapName));
	
	if (!strcmp(MapName, "off"))
	{
		PRINT0("usemap none\n");
	}
	else
	{
		PRINT1("usemap '%s'\n", MapName);
	}

	return true;
}


bool
ParserObj::ParseUseMaterial()
{
	char*			MatName;
	MATCH(LX_TOKEN_USEMATERIAL);
	ASSERT(Lexr->SkipSpaces());
	ASSERT(ParseWord(MatName));
	PRINT1("use material '%s'\n", MatName);

	return true;
}


bool
ParserObj::ParseMaterialLib()
{
	MATCH(LX_TOKEN_MATLIB);
	PRINT0("matlib: ");

	char*			LibName;
	int32			Ln = Lexr->GetLineNumber();
	ASSERT(Lexr->SkipSpaces());
	while (Lexr->GetLineNumber() == Ln)
	{
		char*		LibName;
		ASSERT(ParseWord(LibName));
		_PRINT1("'%s' ", LibName);
		ASSERT(Lexr->SkipSpaces());
	}
	_PRINT0("\n");

	return true;
}


bool
ParserObj::ParseShadow()
{
	char*			FileName;
	MATCH(LX_TOKEN_SHADOW);
	ASSERT(Lexr->SkipSpaces());
	ASSERT(ParseWord(FileName));
	PRINT1("shadow in file '%s' ", FileName);

	return true;
}


bool
ParserObj::ParseTrace()
{
	char*			FileName;
	MATCH(LX_TOKEN_TRACE);
	ASSERT(Lexr->SkipSpaces());
	ASSERT(ParseWord(FileName));
	PRINT1("trace in file '%s' ", FileName);

	return true;
}


bool
ParserObj::ParseCtech()
{
	MATCH(LX_TOKEN_CTECH);

	if (Peek(LX_TOKEN_CPARM))
	{
		float			Reso;
		MATCH(LX_TOKEN_CPARM);
		ASSERT(ParseFloat(Reso));
		PRINT1("ctech cparm, %f\n", Reso);
	}
	else if (Peek(LX_TOKEN_CSPACE))
	{
		float			MaxLen;
		MATCH(LX_TOKEN_CSPACE);
		ASSERT(ParseFloat(MaxLen));
		PRINT1("ctech cspace, %f\n", MaxLen);
	}
	else
	{
		float			MaxDist, MaxAngle;
		MATCH(LX_TOKEN_CURVE);
		ASSERT(ParseFloat(MaxDist));
		ASSERT(ParseFloat(MaxAngle));
		PRINT2("ctech curv, %f %f\n", MaxDist, MaxAngle);
	}

	return true;
}


bool
ParserObj::ParseStech()
{
	MATCH(LX_TOKEN_STECH);

	if (Peek(LX_TOKEN_CPARMA))
	{
		float			ResU, ResV;
		MATCH(LX_TOKEN_CPARMA);
		ASSERT(ParseFloat(ResU));
		ASSERT(ParseFloat(ResV));
		PRINT2("stech parma, %f %f\n", ResU, ResV);
	}
	else if (Peek(LX_TOKEN_CPARMB))
	{
		float			ResUv;
		MATCH(LX_TOKEN_CPARMB);
		ASSERT(ParseFloat(ResUv));
		PRINT1("ctech cparmb, %f\n", ResUv);
	}
	else if (Peek(LX_TOKEN_CSPACE))
	{
		float			MaxLen;
		MATCH(LX_TOKEN_CSPACE);
		ASSERT(ParseFloat(MaxLen));
		PRINT1("stech cspace, %f\n", MaxLen);
	}
	else
	{
		float			MaxDist, MaxAngle;
		MATCH(LX_TOKEN_CURVE);
		ASSERT(ParseFloat(MaxDist));
		ASSERT(ParseFloat(MaxAngle));
		PRINT2("stech curv, %f %f\n", MaxDist, MaxAngle);
	}

	return true;
}


bool
ParserObj::ParseBsp()
{
	int32			Index;
	MATCH(LX_TOKEN_BSP);
	PRINT0("bsp: ");

	for (int32 i = 0; i < 16; i++)
	{
		ASSERT(LexInt(Index));
		_PRINT1("%d ", Index);
	}
	_PRINT0("\n");

	return true;
}


bool
ParserObj::ParseBzp()
{
	int32			Index;
	MATCH(LX_TOKEN_BZP);
	PRINT0("bzp: ");

	for (int32 i = 0; i < 16; i++)
	{
		ASSERT(LexInt(Index));
		_PRINT1("%d ", Index);
	}
	_PRINT0("\n");

	return true;
}


bool
ParserObj::ParseCdc()
{
	int32			Index;
	MATCH(LX_TOKEN_CDC);
	PRINT0("cdc: ");

	while (Peek(LX_TYPE_INTEGER))
	{
		ASSERT(LexInt(Index));
		_PRINT1("%d ", Index);
	}
	_PRINT0("\n");

	return true;
}


bool
ParserObj::ParseCdp()
{
	int32			Index;
	MATCH(LX_TOKEN_CDP);
	PRINT0("cdp: ");

	for (int32 i = 0; i < 16; i++)
	{
		ASSERT(LexInt(Index));
		_PRINT1("%d ", Index);
	}
	_PRINT0("\n");

	return true;
}


bool
ParserObj::ParseRes()
{
	int32			NbSegmU, NbSegmV;
	MATCH(LX_TOKEN_RES);
	ASSERT(LexInt(NbSegmU));
	ASSERT(LexInt(NbSegmV));
	PRINT2("res: %d %d\n", NbSegmU, NbSegmV);

	return true;
}

